#! /usr/bin/python
#
#  AR example version 0.4.0
#  <http://pygame.org/project-AR+example-1622-4243.html>
#
#  An augmented reality and virtual interface example
#  Copyright (C) 2014 slmagik <slmagik@gmail.com>
#
#  This program is free software; you can redistribute it and/or modify
#  it under the terms of the GNU General Public License as published by
#  the Free Software Foundation; either version 2 of the License, or
#  (at your option) any later version.
#
#  This program is distributed in the hope that it will be useful,
#  but WITHOUT ANY WARRANTY; without even the implied warranty of
#  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#  GNU General Public License for more details.
#
#  You should have received a copy of the GNU General Public License along
#  with this program; if not, write to the Free Software Foundation, Inc.,
#  51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
#
#  VideoCapture by Markus Gritsch is under LGPL 2.1
#

import os
import sys
import math
import random
import platform
import pygame as pg
import pygame.font
import pygame.camera
try:
	import psyco
	psyco.full()
except ImportError:
	pass


#Application states
class state:

	CALIBRATION = 0
	TARGETING = 1


#Virtual button
class UIRect:

	def __init__(self, parent, rect, callback, hidden = False,
		color = pg.Color("blue")):
		self.rect = rect
		self.hidden = hidden
		self.waiting = False
		self.color = color
		self.callback = callback
		self.parent = parent
		self.surface = parent.screen
		self.marker = parent.marker
		self.start_time = 0


	def update(self):
		if not self.hidden:

			#Button hover check
			if self.marker.xy[0] - self.rect[2] < self.rect[0] \
				< self.marker.xy[0] + self.marker.size and self.marker.xy[1] \
				- self.rect[3] < self.rect[1] < self.marker.xy[1] \
				+ self.marker.size:

				pg.draw.rect(self.surface, self.color, self.rect, 0)

				#Start timing
				if not self.waiting:
					self.waiting = True
					self.start_time = pg.time.get_ticks()

				#Button hover event
				elif pg.time.get_ticks() - self.start_time > 750:
						self.waiting = False
						self.callback(self)

			#Cancel timing
			else:
				self.waiting = False
				pg.draw.rect(self.surface, self.color, self.rect, 3)


#Target object marker & color-chooser
class Marker:

	def __init__(self, surface):
		self.size = 40
		self.hidden = False
		self.surface = surface
		self.surf_size = surface.get_size()

		#Color-chooser defaults
		self.filter_mode = 1
		self.choose_color = True
		self.color = (230, 230, 230)
		self.threshold = [25, 25, 25]

		#Center position
		self.xy = [(self.surf_size[0] - self.size) / 2,
			(self.surf_size[1] - self.size) / 2]
		self_half = self.size / 2
		self.centroid = [self.xy[0] + self_half, self.xy[1] + self_half]


	#Blitting
	def update(self):
		if not self.hidden:
			pg.draw.rect(self.surface, pg.Color("red"), (self.xy,
				(self.size, self.size)), 3)


class Ball:

	def __init__(self, surface, targets, marker, image):
		self.size = 40
		self.marker = marker
		self.targets = targets
		self.targets.append(self)
		self.surface = surface
		self.surf_size = surface.get_size()

		#Ball sprite
		self.image = pg.transform.smoothscale(image,
			(self.size, self.size)).convert_alpha()

		#Random vector & position
		self.speed = [3 * random.choice((-1, 1)), 3 * random.choice((-1, 1))]
		self.xy = [random.randint(0, self.surf_size[0] - self.size),
			random.randint(0, self.surf_size[1] - self.size)]


	def _move(self):

		#X coord check
		if self.xy[0] < 0:
			self.xy[0] = 0
			self.speed[0] = -self.speed[0]

		elif self.xy[0] > self.surf_size[0] - self.size:
			self.xy[0] = self.surf_size[0] - self.size
			self.speed[0] = -self.speed[0]

		#Y coord check
		if self.xy[1] < 0:
			self.xy[1] = 0
			self.speed[1] = -self.speed[1]

		elif self.xy[1] > self.surf_size[1] - self.size:
			self.xy[1] = self.surf_size[1] - self.size
			self.speed[1] = -self.speed[1]

		#Marker collision check
		if self.marker.xy[0] - self.size < self.xy[0] < self.marker.xy[0] \
			+ self.marker.size and self.marker.xy[1] - self.size \
			< self.xy[1] < self.marker.xy[1] + self.marker.size:

			centroid = self.get_centroid()
			centroid_diff = (centroid[0] - self.marker.centroid[0],
				centroid[1] - self.marker.centroid[1])

			#NW direction
			if self.speed[0] < 0 and self.speed[1] < 0:

				#NW quadrant
				if centroid_diff[0] <= 0 and centroid_diff[1] <= 0:
					self.xy[0] += self.speed[0]
					self.xy[1] += self.speed[1]

				#NE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] <= 0:
					self.speed[0] = -self.speed[0]

				#SW quadrant
				elif centroid_diff[0] <= 0 and centroid_diff[1] > 0:
					self.speed[1] = - self.speed[1]

				#SE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] > 0:
					if abs(centroid_diff[0]) <= abs(centroid_diff[1]):
						self.speed[1] = -self.speed[1]
					else:
						self.speed[0] = -self.speed[0]

			#NE direction
			elif self.speed[0] > 0 and self.speed[1] < 0:

				#NW quadrant
				if centroid_diff[0] <= 0 and centroid_diff[1] <= 0:
					self.speed[0] = -self.speed[0]

				#NE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] <= 0:
					self.xy[0] += self.speed[0]
					self.xy[1] += self.speed[1]

				#SW quadrant
				elif centroid_diff[0] <= 0 and centroid_diff[1] > 0:
					if abs(centroid_diff[0]) <= abs(centroid_diff[1]):
						self.speed[1] = -self.speed[1]
					else:
						self.speed[0] = -self.speed[0]

				#SE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] > 0:
					self.speed[1] = -self.speed[1]

			#SW direction
			elif self.speed[0] < 0 and self.speed[1] > 0:

				#NW quadrant
				if centroid_diff[0] <= 0 and centroid_diff[1] <= 0:
					self.speed[1] = -self.speed[1]

				#NE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] <= 0:
					if abs(centroid_diff[0]) <= abs(centroid_diff[1]):
						self.speed[1] = -self.speed[1]
					else:
						self.speed[0] = -self.speed[0]

				#SW quadrant
				elif centroid_diff[0] <= 0 and centroid_diff[1] > 0:
					self.xy[0] += self.speed[0]
					self.xy[1] += self.speed[1]

				#SE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] > 0:
					self.speed[0] = -self.speed[0]

			#SE direction
			elif self.speed[0] > 0 and self.speed[1] > 0:

				#NW quadrant
				if centroid_diff[0] <= 0 and centroid_diff[1] <= 0:
					if abs(centroid_diff[0]) <= abs(centroid_diff[1]):
						self.speed[1] = -self.speed[1]
					else:
						self.speed[0] = -self.speed[0]

				#NE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] <= 0:
					self.speed[1] = -self.speed[1]

				#SW quadrant
				elif centroid_diff[0] <= 0 and centroid_diff[1] > 0:
					self.speed[0] = - self.speed[0]

				#SE quadrant
				elif centroid_diff[0] > 0 and centroid_diff[1] > 0:
					self.xy[0] += self.speed[0]
					self.xy[1] += self.speed[1]

		self.xy[0] += self.speed[0]
		self.xy[1] += self.speed[1]


	#Moving & blitting
	def update(self):
		self._move()
		self.surface.blit(self.image, self.xy)


	def get_centroid(self):
		self_half = self.size / 2
		return (self.xy[0] + self_half, self.xy[1] + self_half)


class App:

	def __init__(self):
		pg.init()
		pg.font.init()
		pg.camera.init()
		random.seed()

		self.state = state.CALIBRATION
		self.size = (640, 480)
		self.title = "AR example"
		self.screen = None
		self.running = False
		self.marker = None
		self.filter = 0
		self.image = pg.image.load("data/ball.png")
		self.sprites = []
		font_name = pg.font.get_default_font()
		self.font = pg.font.SysFont(font_name, 40)
		self.clock = pg.time.Clock()

		#Camera init
		cams = pygame.camera.list_cameras()
		if cams:
			self.camera = pg.camera.Camera(cams[0], self.size, "RGB")
			self.camera.start()


	#Exit button method
	def _exit(self, caller):
		caller.hidden = True
		caller.parent.confirm_button.hidden = False
		caller.parent.cancel_button.hidden = False


	#Cancel button method
	def _cancel_exit(self, caller):
		caller.parent.exit_button.hidden = False
		caller.parent.confirm_button.hidden = True
		caller.hidden = True


	#Confirm button method
	def _confirm_exit(self, caller):
		caller.parent.running = False


	def run(self):

		#Display init
		self.screen = pg.display.set_mode(self.size,
			pg.DOUBLEBUF | pg.HWSURFACE)
		pg.display.set_caption(self.title)
		pygame.mouse.set_visible(False)
		pg.key.set_repeat(200, 50)
		self.snapshot = pg.Surface(self.size).convert()

		#Sprites init
		self.marker = Marker(self.screen)
		for i in xrange(5):
			Ball(self.screen, self.sprites, self.marker, self.image)

		#Buttons
		self.exit_button = UIRect(self, (20, 20, 100, 50), self._exit)
		self.cancel_button = UIRect(self, ((self.size[0] - 200) / 2,
			(self.size[1] - 50) / 2, 100, 50), self._cancel_exit,
			True, pg.Color("green"))
		self.confirm_button = UIRect(self, ((self.size[0] + 100) / 2,
			(self.size[1] - 50) / 2, 100, 50), self._confirm_exit,
			True, pg.Color("red"))

		#Main loop
		self.running = True
		while self.running:

			#Blitting
			self.update()

			#Events
			for event in pg.event.get():
				if event.type == pg.KEYDOWN:

					#Targeting state
					if self.state == state.TARGETING:
						if event.key == pg.K_MINUS:
							if self.sprites:
								self.sprites.pop()

						elif event.key == pg.K_EQUALS:
							Ball(self.screen, self.sprites, self.marker,
								self.image)

						elif event.key == pg.K_1:
							self.marker.hidden = not self.marker.hidden

						elif event.key == pg.K_RETURN:
							self.marker.hidden = True
							self.marker.size = 40
							self.marker.filter_mode = 0
							self.marker.choose_color = True
							self.marker.xy = [(self.size[0] \
								- self.marker.size) / 2,
								(self.size[1] - self.marker.size) / 2]
							self.state = not self.state

					#Calibration state
					elif self.state == state.CALIBRATION:
						if event.key == pg.K_MINUS:
							if self.marker.size > 10:
								self.marker.size -= 10
							else:
								self.marker.size = 10

						elif event.key == pg.K_EQUALS:
							if self.marker.size < 100:
								self.marker.size += 10
							else:
								self.marker.size = 100

						elif event.key == pg.K_LEFT:
							self.marker.xy[0] -= 10
							if self.marker.xy[0] < 0:
								self.marker.xy[0] = 0

						elif event.key == pg.K_RIGHT:
							self.marker.xy[0] += 10
							if self.marker.xy[0] > (self.size[0] \
								- self.marker.size):
								self.marker.xy[0] = self.size[0] \
									- self.marker.size

						elif event.key == pg.K_UP:
							self.marker.xy[1] -= 10
							if self.marker.xy[1] < 0:
								self.marker.xy[1] = 0

						elif event.key == pg.K_DOWN:
							self.marker.xy[1] += 10
							if self.marker.xy[1] > (self.size[1] \
								- self.marker.size):
								self.marker.xy[1] = self.size[1] \
									- self.marker.size

						elif event.key == pg.K_RETURN:
							self.state = not self.state
							self.marker.hidden = False

						elif event.key == pg.K_SPACE:
							self.marker.choose_color = \
								not self.marker.choose_color

						elif event.key == pg.K_1:
							self.marker.filter_mode = 0

						elif event.key == pg.K_2:
							self.marker.filter_mode = 2

						elif event.key == pg.K_3:
							self.marker.filter_mode = 1

						elif event.key == pg.K_q:
							self.marker.threshold[0] += 1
							if self.marker.threshold[0] > 255:
								self.marker.threshold[0] = 255

							elif self.marker.threshold[0] < 0:
								self.marker.threshold[0] = 0

						elif event.key == pg.K_w:
							self.marker.threshold[1] += 1
							if self.marker.threshold[1] > 255:
								self.marker.threshold[1] = 255

							elif self.marker.threshold[1] < 0:
								self.marker.threshold[1] = 0

						elif event.key == pg.K_e:
							self.marker.threshold[2] += 1
							if self.marker.threshold[2] > 255:
								self.marker.threshold[2] = 255

							elif self.marker.threshold[2] < 0:
								self.marker.threshold[2] = 0

						elif event.key == pg.K_a:
							self.marker.threshold[0] -= 1
							if self.marker.threshold[0] > 255:
								self.marker.threshold[0] = 255

							elif self.marker.threshold[0] < 0:
								self.marker.threshold[0] = 0

						elif event.key == pg.K_s:
							self.marker.threshold[1] -= 1
							if self.marker.threshold[1] > 255:
								self.marker.threshold[1] = 255

							elif self.marker.threshold[1] < 0:
								self.marker.threshold[1] = 0

						elif event.key == pg.K_d:
							self.marker.threshold[2] -= 1
							if self.marker.threshold[2] > 255:
								self.marker.threshold[2] = 255

							elif self.marker.threshold[2] < 0:
								self.marker.threshold[2] = 0

					if event.key == pg.K_ESCAPE:
						self.running = False

				elif event.type == pg.QUIT:
					self.running = False


	#Blitting
	def update(self):

		#FPS limit
		self.clock.tick(60)

		#Camera snapshot
		try:
			self.snapshot = pygame.transform.flip(
					self.camera.get_image(self.snapshot),
						True, False).convert()
		except:
			self.snapshot = pygame.transform.flip(
					self.camera.get_image(self.snapshot),
						True, False).convert()

		#Targeting state
		if self.state == state.TARGETING:
			self.screen.blit(self.snapshot, (0, 0))
			for sprite in self.sprites:
				sprite.update()

			#Update buttons
			self.exit_button.update()
			self.confirm_button.update()
			self.cancel_button.update()

			#Track object
			mask = pygame.mask.from_threshold(self.snapshot,
				self.marker.color, self.marker.threshold)
			mask_size = mask.count()
			self.marker.size = math.sqrt(mask_size) * 1.25
			if mask_size > 300:
				self.marker.centroid = mask.centroid()
				marker_half = self.marker.size / 2
				self.marker.xy = [self.marker.centroid[0] - marker_half,
					self.marker.centroid[1] - marker_half]
				self.marker.update()
			else:
				self.marker.xy = self.size

		#Calibration state
		elif self.state == state.CALIBRATION:
			if self.marker.choose_color:
				self.marker.color = pg.transform.average_color(self.snapshot,
					(self.marker.xy, (self.marker.size, self.marker.size)))

			#Mask color
			pygame.transform.threshold(self.screen, self.snapshot,
				self.marker.color, self.marker.threshold, pg.Color("black"),
					self.marker.filter_mode)
			if not self.marker.filter_mode:
				self.screen.blit(self.snapshot, (0, 0))

			#Show color-chooser
			self.marker.update()
			pg.draw.rect(self.screen, self.marker.color, (0, 0, 100, 100), 0)

		#Show FPS
		fps_count = str(int(self.clock.get_fps()))
		fps = self.font.render(fps_count, True, pg.Color("white"))
		self.screen.blit(fps, (self.size[0] - fps.get_size()[0], 0))

		pg.display.flip()